unit Powerfrequency;
//Empfngt ber RS232 Periodendaten vom Netzfrequenzmonitor
//berechent die Netzfrequenz und zeigt den zeitlichen Verlauf
//in einem Diagramm an

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, CommInt, ExtCtrls, GrHilf_U, Spin, ExtDlgs, ComCtrls, IniFiles,
  Zeit_U;

type
  TForm1 = class(TForm)
    Comm1: TComm;
    Image1: TImage;
    Panel1: TPanel;
    EditFrequenz: TEdit;
    EditRxData: TEdit;
    Label3: TLabel;
    Label1: TLabel;
    EditStartZeit: TEdit;
    Label7: TLabel;
    EditXZeit: TEdit;
    Label2: TLabel;
    SpinEditXMax: TSpinEdit;
    Label4: TLabel;
    Label6: TLabel;
    Speichern: TButton;
    StartButt: TButton;
    SaveDialog1: TSaveDialog;
    ComboPort: TComboBox;
    RadioSeriell: TRadioButton;
    RadioDatei: TRadioButton;
    OpenDialog1: TOpenDialog;
    SpinOffset: TSpinEdit;
    Label5: TLabel;
    Label8: TLabel;
    EditDatum: TEdit;
    Label9: TLabel;
    procedure Comm1RxChar(Sender: TObject; Count: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure SpinEditXMaxChange(Sender: TObject);
    procedure StartButtClick(Sender: TObject);
    procedure SpeichernClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SetValues;
    procedure BildZeichnen;
    procedure SpinOffsetChange(Sender: TObject);
    procedure StartZeitLesen;
    procedure FileOpen();
    procedure FileClose();
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
  end;

const
  //Messzeit des Mikrocontrollers 12*Quarzperiode
  TIMEBASE = 12.0/12000000.0;   // 12MHz Quarz
  //TIMEBASE = 12.0/22118400.0;  // 22.1184< MHz Quarz

var
  Form1: TForm1;
  LineData : String;       //Empfangsdaten der ser. Schnittstelle
  Bitmap1 : TBitmap;
  X : Real;                //Zeit fr aktuellen Mepunkt
  AltDate : TDateTime;     //Zum Erkennen des Datumswechsels ntig
  T0   : TDateTime;        //Startzeitpunkt des Diagramms
  StartDatum : TDateTime;  //Wird in das Bitmap beim speichern eingetragen
  StartZeit  : TDateTime;
  Xmin : Real;             //X-Achsenwerte des Diagramms
  Xmax : Real;
  Ymin : Real;             //Y-Achsenwerte des Diagramms
  Ymax : Real;
  BmpExists : Boolean;
  Freq : Double;           //Netzfrequenz
  AltFreq : Double;        //Frequenz der letzten Sekunde
  Progstart : Boolean;     //Flag "Programm neu gestartet"
  StartStd  : Integer;     //Ganze Stunde, damit das Diagramm bei vollen Stunden beginnt
  IniDir : String;         //Programmverzeichnis
  NetzfreqIni: TIniFile;
  F: TextFile;             //File mit Messwerten
  OffH : Integer;          //Offset [h]
  FileGewaehlt : Boolean;  //Flag "File zum ffnen ausgewhlt"
  FileName : String;
  FWriteOpen : Boolean;    //Flag "File zum schreiben geffnet"
  StartZeitGelesen : Boolean; //Fr manuelle Eingabe des Startzeitpunkts bentigt
  Warten : Boolean;
  LineCount : Longint;

implementation


{$R *.DFM}

procedure TForm1.Comm1RxChar(Sender: TObject; Count: Integer);
//Seriell empfangene Zeichen auswerten.
//gesendet wird ein fnfstelliger Wert +Cr+LF: XXXXX,CR,LF
//(Periodenzeit in Mikrosekunden)
//Nach Empfang eines Wertes Digramm zeichenen und Wert in Datei schreiben

label
  Fehler;

type
  CharBuf = array[0..4096] of Char;

var
  Buffer: ^CharBuf;
  Bytes, P: Integer;
  TStr : String;
  PerCount : Integer;

begin
  GetMem(Buffer, Comm1.ReadBufSize);
  try
    If Count>Comm1.ReadbufSize then exit;
    Fillchar(Buffer^, Comm1.ReadBufSize, 0); //Puffer lschen
    Bytes := Comm1.Read(Buffer^, Count);     //Zeichen lesen
    if Bytes = -1 then
      ShowMessage('Error reading incoming data...')
    else
    begin
      //Zeichen Byteweise auswerten, da ein Telgramm in mehreren
      //Blcken ankommen kann. Endekennzeichnung einer Zeile ist LF (#10).
      for P := 0 to Bytes - 1 do
      begin
        case Buffer^[P] of
          #13: ;     //Keine Aktion bei CR
          #10: begin //LF ist Endekennzeichnung einer Zeile
                 //Es mssen 5 Zeichen empfangen werden (CR,LF zhlen nicht)
                 If Length(LineData)<>5
                   then
                     begin
                       TStr:='';
                       LineData:='';
                     end
                   else
                     begin
                       if AltDate<>Date then
                         begin
                           FileClose;
                           FileOpen;
                         end;
                       AltDate:=Date;
                       TStr:=Trim(LineData);
                       LineData:='';
                       //Aktuelle Zeit + empf. Wert anzeigen
                       EditRxData.Text:=TimeToStr(Time)+' '+TStr;
                       X:=(Date+Time-T0)*24; //Zeit in Stunden
                       EditXZeit.Text:=FloatToStrF(X,ffFixed,8,4);
                       PerCount:=StrToInt(TStr);
                       Freq:=1/(PerCount*TIMEBASE);
                       EditFrequenz.Text:=FloatToStrF(Freq,ffFixed,7,3);
                       //Fehlmessung durch "Spratzler" verhindern
                       //Neuer Wert mu Alterwert +- 1Hz sein
                       If Abs(Freq-AltFreq)<1 then
                         begin
                           //Wert anzeigen
                           //Zum ersten Punkt keine Linie zeichnen
                           If Progstart then
                             begin
                               GrHilf._MoveTo(Bitmap1,X,Freq);
                               Progstart:=False;
                             end
                           else GrHilf._LineTo(Bitmap1,X,Freq);
                           Image1.Canvas.Draw(0,0,Bitmap1);
                           Writeln(F,TStr);
                         end;
                       AltFreq:=Freq;
                 end; // If Length(Linedata)...
               end // #10:
           else
             begin
               //Byte an Empfangsdaten anfgen
               LineData := LineData + CharBuf(Buffer^)[P];
               //"Notausstieg"
               If Length(LineData)>16 then
                 begin
                   LineData:=''; //Fehler
                 end;
             end;
        end; //case
      end; //for do
    end;
  Application.ProcessMessages;
  finally
    FreeMem(Buffer);
  end;
end;

procedure TForm1.SetValues;
//Wertbereich des Bitmaps definieren

begin
  XMin:=0;                               //XMax wird duch einen SpinEdit eingegeben
  If BmpExists then Bitmap1.Free;        //Eventuell altes Bitmap freigeben
  Bitmap1:=TBitmap.Create;               //Bitmap erzeugen
  BmpExists:=True;
  Bitmap1.Width:=Image1.Width;
  Bitmap1.Height:=Image1.Height;
  Bitmap1.PixelFormat:=pf4bit;
  //"Weltgrenzen" fr die Anzeige setzen, die Werte werden
  //dann in Pixelwerte fr die Anzeige umgerechent
  GrHilf.SetAereaLimit(Xmin,Ymin,Xmax,Ymax);
  //Freirume zur Beschriftung der Achsen setzen
  GrHilf.SetBmpLimit(30,0,Bitmap1.Width,Bitmap1.Height-20);
  GrHilf.XGitter(Bitmap1,Xmin,Xmax,1,Ymin,Ymax);
  GrHilf.YGitter(Bitmap1,Ymin,Ymax,0.1,Xmin,Xmax);
  GrHilf.YLabel(Bitmap1,Ymin,Ymax,0.5,0);
  GrHilf.YLabel_Klein(Bitmap1,Ymin,Ymax,0.1,0);
  GrHilf.XLabel(Bitmap1,Xmin,Xmax,5,StartStd);
  GrHilf.XLabel_Klein(Bitmap1,Xmin,Xmax,1,StartStd);
  Bitmap1.Canvas.Pen.Color:=clGreen;
  //50Hz +- 20mHz Linien zeichnen
  GrHilf._MoveTo(Bitmap1,XMin,50.02);
  GrHilf._LineTo(Bitmap1,Xmax,50.02);
  GrHilf._MoveTo(Bitmap1,XMin,49.98);
  GrHilf._LineTo(Bitmap1,Xmax,49.98);
  GrHilf._MoveTo(Bitmap1,0,0);
  Bitmap1.Canvas.Pen.Color:=clRed;
end;


procedure TForm1.FormCreate(Sender: TObject);
//Einstellungen und Fenterposition aus Ini-File lesen

begin
  IniDir:=ExtractFilePath(Application.ExeName);
  NetzFreqIni := TIniFile.Create(IniDir+'Powerfrequency.Ini');
  with NetzFreqIni do
    begin
     ComboPort.ItemIndex:=ReadInteger('Com','Port',0);
     SpinEditXMax.Value:=ReadInteger('Values','XMax',2);
     Form1.Top:=ReadInteger('Values','Top',10);
     Form1.Left:=ReadInteger('Values','Left',10);
     Form1.Width:=ReadInteger('Values','Width',600);
     Form1.Height:=ReadInteger('Values','Height',300);
     RadioSeriell.Checked:=ReadBool('Values','Serial',True);
     RadioDatei.Checked:=ReadBool('Values','File',False);
     YMin:=ReadFloat('Values','YMin',49.5);
     YMax:=ReadFloat('Values','YMax',50.5);
    end;
  NetzFreqIni.Free;
  XMax:=SpinEditXMax.Value;  //kann vom User eingestellt werden
  BmpExists:=False;
  FileGewaehlt:=False;
  FWriteOpen:=False;
  StartZeitGelesen:=False;
  Warten:=False;
  {-----------------------------------------------------------------------
    Ist im ShortDateFormat das Trennzeichen '/' so bestimmt der
    DateSeparator das Datumsformat, alle anderen Zeichen werden direkt
    in das Datumsformat bernommen.
   ------------------------------------------------------------------------}
  //Diese Varibalen bestimmen das Anzeigeformat von Datum und Uhrzeit

  DateSeparator:='.';
  ShortDateFormat:='dd/mm/yyyy'; //Der DateSeparator bestimmt das Ausgabeformat
//  ShortDateFormat:='ddxmmxyyyy'; //ShortDatFomat bestimmt das Trennzeichen

  TimeSeparator:=':';
  LongTimeFormat:='hh:mm:ss';
  
//  ShowMessage(ShortDateFormat+' '+DateSeparator+' '+ShortTimeFormat+' '+LongTimeFormat+' '+TimeSeparator);
//  ShowMessage(DateToStr(Now)+' '+TimeToStr(Now));

end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);

begin
  If Comm1.Enabled then Comm1.Close;
  Bitmap1.Free;
end;

procedure TForm1.SpinEditXMaxChange(Sender: TObject);
//Xmax hat sich durch Benutzereingabe gendert
begin
  XMax:=SpinEditXMax.Value;
  If RadioDatei.Checked then
    begin
      ProgStart:=True;
      //Bei schnell wiederholten Aufrufen
      //verhindern, dass das Datenfile 2x geffnet wird.
      //Warten=True, solange die Bildberechnung luft
      If not Warten then
        begin
         StartZeitLesen;
         BildZeichnen;
        end;
    end;
end;

procedure TForm1.BildZeichnen;
//Zeichnet das Bitmap mit den Messwereten

var
  TmpStr: String;
  Count: Integer;      //Zhlt die glesen Messwerte (1Messwert-> 1Sekunde)
  PerCount : Integer;

begin
  //kein File mit Daten gewhlt, nicht zeichnen
  If Not FileGewaehlt then Exit;
  EditStartZeit.Text:=TimeToStr(StartZeit);
  //Diagramm immer mit einer vollen Stunde beginnen
  StartStd:=Trunc(24.0*StartZeit+OffH);
  //Datum neu berechnen falls sich durch den Offset, das Datum gendert hat
  EditDatum.Text:=DateToStr(StartDatum+StartStd/24);
  SetValues;                                  //Diagrammwerte setzen
  Count:=0;
  while not EOF(F) do
    begin
     Readln(F, TmpStr);
     LineCount:=LineCount+1;
     TmpStr:=Trim(TmpStr);
     If Length(TmpStr)=5 then
      begin
         X:=StartZeit*24-StartStd+Count/3600;  //jede Sekunde ein Mewert
         PerCount:=StrToInt(TmpStr);
         Freq:=1/(PerCount*TIMEBASE);
         //Fehlmessung durch "Spratzler" verhindern
         If Abs(Freq-AltFreq)<1 then
           begin
             //Zum ersten Punkt keine Linie zeichnen
             If Progstart then
               begin
                 GrHilf._MoveTo(Bitmap1,X,Freq);
                 Progstart:=False;
               end
             else GrHilf._LineTo(Bitmap1,X,Freq);
           end;
         AltFreq:=Freq;
         Count:=Count+1;
       end //If Length..
     else
       //Eintrag hat die falsche Lnge
       begin
         ShowMessage('Error in line '+IntToStr(LineCount)+': '
                      +#10+#13+TmpStr);
       end;
     Application.ProcessMessages;
    end;
  CloseFile(F);
  Image1.Canvas.Draw(0,0,Bitmap1);
  Warten:=False;
end;

procedure TForm1.StartZeitLesen;
//ffnet die Dati, liest (falls vorhanden) die Startzeit aus der Datei.
//Falls die Zeit nicht in der Datei steht, wird eine Eingabe angefordert

var
  TmpStr:String;
  DatumIO : Boolean;

begin
   If not FileGewaehlt then Exit;
   Warten:=True;
   AssignFile(F, OpenDialog1.FileName);   { Datei auswhlen }
   Reset(F);
   Readln(F,TmpStr);
   LineCount:=LineCount+1;
   Trim(TmpStr);
   DatumIO:=False;
   //1. Zeile enthlt Datum und Startzeit der Aufzeichnungen
   If Length(TmpStr)=19 then
     begin
       //Lnge der 1. Zeile ist i.O. ('tt.mm.jjjj hh.mm.ss')
       DatumIO:=True;
       //Prfen, ob 1. Zeile Datum in der Form tt.mm.jjjj hh:mm:ss ist
       If TmpStr[3]<>'.' then DatumIO:=False;
       If TmpStr[6]<>'.' then DatumIO:=False;
       If TmpStr[14]<>':' then DatumIO:=False;
       If TmpStr[17]<>':' then DatumIO:=False;
       If Length(TmpStr)<>19 then DatumIO:=False;
     end;
   If DatumIO then //1. Zeile enthlt Datum und Starttzeit
     begin
      StartDatum:=StrToDate(Copy(TmpStr,1,10));
      StartZeit:=StrToTime(Copy(TmpStr,12,8));
     end
   else
     begin
      Reset(F); //Lesezeiger wieder auf Anfang
      If StartZeitGelesen then Exit;
      //Fenster zur Eingabe von Startzeit und Offset ffnen
      Form2.ShowModal;
      StartDatum:=StrToDate(Form2.EditDatum.Text);    //Datum
      StartZeit:=StrToTime(Form2.EditStartZeit.Text); //Uhrzeit
      StartZeitGelesen:=True;
     end;
end;

procedure TForm1.FileOpen();

var
  Uhrzeit : String;
  Datum : String;

begin
  Uhrzeit:=TimeToStr(Now);
  Datum:=DateToStr(Now);
  //Filenamen in der Form 'NetzLog_jjjjmmtt_hhmmss.txt' erzeugen
  FileName:='NetzLog_'
            +Copy(Datum,7,4)+Copy(Datum,4,2)+Copy(Datum,1,2)+'_'
            +Copy(Uhrzeit,1,2)+Copy(UhrZeit,4,2)+Copy(UhrZeit,7,2)+'.txt';
  StartDatum:=Date;            //Aktuelles Datum
  StartZeit:=Time;
  EditStartZeit.Text:=TimeToStr(Time);
  //Als Startstunde eine volle Stunde setzen
  StartStd:=StrToInt(Copy(Uhrzeit,1,2)); //Die ersten 2 Zeichen sind die Stunde
  T0:=Date+StartStd/24;
  SetValues;
  EditDatum.Text:=DateToStr(StartDatum);
  AssignFile(F,FileName);
  ReWrite(F);
  //Als erste Zeile Datum und Startzeit schreiben
  Writeln(F,DateToStr(Now)+' '+TimeToStr(Now));
  FWriteOpen:=True;
  Progstart:=True;
end;

procedure TForm1.FileClose();

begin
 //Falls Datei zum schreiben geffnet ist, diese schliesen
 If FWriteOpen then
  begin
   Flush(F);
   CloseFile(F);
   FWriteOpen:=False;
  end;
end;


procedure TForm1.StartButtClick(Sender: TObject);

var
  Uhrzeit : String;

begin
 If RadioSeriell.Checked then
   //Daten werden ber serielle schnittstelle empfangen
   begin
    If not Comm1.Enabled then
     //File anlegen, Comport ffnen, Messung starten
     begin
      FileOpen;                      //File mit Messwerten ffnen
      Comm1.DeviceName:=ComboPort.Text;
      Comm1.Open;
      X:=0.;
      StartButt.Caption:='Stop';

     end
    else
     begin
       //Compot schlieen, File schlieen
       Comm1.Close;
       StartButt.Caption:='Start';
       FileClose;
     end;
   end;

  If RadioDatei.checked then
   //Werte werden aus einer Datei gelsen
   begin
    if OpenDialog1.Execute then          { Dialog zum Dateiffnen anzeigen }
     begin
       LineCount:=0;
       FileGewaehlt:=True;
       ProgStart:=True;
       StartZeitGelesen:=False;
       StartZeitLesen;
       OffH:=SpinOffset.Value;
       BildZeichnen;
    end;

   end;
end;


procedure TForm1.SpeichernClick(Sender: TObject);
//Datum eintragen und Bitmap speichern

var
  FileName:String;
  Ext:String;
begin
 If SaveDialog1.Execute then
  begin
   //Falls das Datum durch die Offsetstunden gewechselt hat, neuen Tag berechnen
   //und Datum ins Bitmap schreiben
   Bitmap1.Canvas.TextOut(32,5, 'Datum: '+DateToStr(StartDatum+(StartStd/24)));
   FileName:=SaveDialog1.Filename;
   Ext:=ExtractFileExt(FileName);
   If Ext<>'.bmp' then FileName:=ExtractFileName(FileName)+'.bmp';
   if FileExists(FileName) then
     if MessageDlg('Wollen Sie ' + ExtractFileName(FileName) +
        ' berschreiben?', mtConfirmation, [mbYes, mbNo], 0) <> IDYes then  Exit;
   Bitmap1.SaveToFile(FileName);
   ProgStart:=True; //MoveTo beim nchsten Messwert, statt LineTo
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
//Alle Einstellungen im Ini-File speichern

begin
  If FWriteOpen then
   begin
    Flush(F);
    CloseFile(F);
   end;
  NetzFreqIni := TIniFile.Create(IniDir+'Powerfrequency.Ini');
  with NetzFreqIni do
    begin
     WriteInteger('Com','Port',ComboPort.ItemIndex);
     WriteInteger('Values','XMax',SpinEditXMax.Value);
     WriteBool('Values','Serial',RadioSeriell.Checked);
     WriteBool('Values','File',RadioDatei.Checked);
     WriteInteger('Values','Top',Top);
     WriteInteger('Values','Left',Left);
     WriteInteger('Values','Width',Width);
     WriteInteger('Values','Height',Height);
     WriteFloat('Values','YMin',YMin);
     WriteFloat('Values','YMax',YMax);
    end;
  NetzFreqIni.Free;
end;

procedure TForm1.SpinOffsetChange(Sender: TObject);
//Benutzer hat den Offset gendert, Bitmap neu zeichnen

begin
  If RadioDatei.Checked then
    begin
      ProgStart:=True;
      OffH:=SpinOffset.Value;
      //Bei schnell wiederholten Aufrufen
      //verhindern, dass das Datenfile 2x geffnet wird
      //Warten=True, solange die Bildberechnung luft
      If not Warten then
        begin
         StartZeitLesen;
         BildZeichnen;
        end;
    end;
end;

end.
